#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <plib.h>
#include <xc.h>
#include "Config.h"
#include "pt_cornell_1_2_1.h"
// graphics libraries
#include "tft_master.h"
#include "tft_gfx.h"
#include "snare.h"


// === 16:16 fixed point macros === //
typedef signed int fix16 ;
#define multfix16(a,b) ((fix16)(((( signed long long)(a))*(( signed long long)(b)))>>16)) //multiply two fixed 16:16
#define float2fix16(a) ((fix16)((a)*65536.0)) // 2^16
#define fix2float16(a) ((float)(a)/65536.0)
#define fix2int16(a)    ((int)((a)>>16))
#define int2fix16(a)    ((fix16)((a)<<16))
#define divfix16(a,b) ((fix16)((((signed long long)(a)<<16)/(b)))) 
#define sqrtfix16(a) (float2fix16(sqrt(fix2float16(a)))) 
#define absfix16(a) abs(a)
#define onefix16 0x00010000 // int2fix16(1)
#define two32 4294967296.0 // 2^32

// === SYNTHESIS SETUP === //
#define Fs 20000.0
#define sine_table_size 256

// === DAC ==== //
#define DAC_config_chan_A 0b0011000000000000
#define DAC_config_chan_B 0b1011000000000000

// === Keypad Debounce FSM ==========================================
#define release 0
#define mayPressed 1
#define pressed 2
#define mayRelease 3
#define sustain_constant float2fix16(256.0/20000.0) ; // seconds per decay update

// === DAC ===//
volatile unsigned int DAC_data;

// === FM Synthesis === //
volatile unsigned int phase_accum_fm, phase_incr_fm=1.8*261.0*two32/Fs ;// 
volatile unsigned int phase_accum_main, phase_incr_main=261.0*two32/Fs ;//
volatile fix16 sine_table[sine_table_size];
volatile unsigned int random_table[sine_table_size];
volatile fix16 env_fm;
volatile fix16 dk_fm=float2fix16(0.98), attack_fm=float2fix16(0.0);
volatile fix16 dk_state_fm, attack_state_fm;
volatile fix16 env_main, wave_main;
volatile fix16 dk_main=float2fix16(0.995), attack_main=float2fix16(0.0);
volatile fix16 dk_state_main, attack_state_main;
volatile fix16 fm_depth=float2fix16(1.75);
volatile int dk_interval;
volatile fix16 sustain_state=0, sustain_interval=0;
volatile int selectFM = 0, selectKS = 0;
volatile int p = 0, b = 0, s = 0, a = 0, idx = 0, previdx = 0, sample = 0; 

// === Karplus-Strong Synthesis === //
//volatile int p;
//volatile fix16 s, d;
//volatile fix16 sample[];
//volatile int selectKS = 0;
// === ISR === //
volatile int isr_time;

// === Keypad + Play Control=== //
//0-9, A-D, * and #
static int keytable[16] = {0x088, 0x21, 0x81, 0x101, 0x22, 0x82, 0x102, 0x24, 0x84, 0x104, 0x201, 0x202, 0x204, 0x208, 0x28, 0x108};
static int keyString[9] = {-1, -1, -1, -1, -1, -1, -1, -1, -1}; 
static int instrString[10] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1};
volatile int C_length = 76;
volatile int D_length = 68;
volatile int E_length = 60;
volatile int F_length = 57;
volatile int G_length = 51;
volatile int A_length = 45;
volatile int B_length = 40;
volatile int C5_length = 38;
volatile int D5_length = 34;

volatile signed int stringC[76];
volatile signed int stringD[68];
volatile signed int stringE[60];
volatile signed int stringF[57];
volatile signed int stringG[51];
volatile signed int stringA[45];
volatile signed int stringB[40];
volatile signed int stringC5[38];
volatile signed int stringD5[34];
volatile int key;

volatile int playString[50];
volatile int playIndiv = 0, playBack = 0;
volatile int selectSnare = 0;
// === SPI === //
static int adc_reading;
volatile SpiChannel spiChn = SPI_CHANNEL2;  //SPI Channel to use
volatile int spiClkDiv = 2; //20 MHz speed for this RAM
volatile int index, upperIndex, NumISR, packed, DAC_input, i, j, k, counter, randval;
// === Protothreads === //
static struct pt pt_keypad, pt_tick;

// === TFT === //
static char display[80];


void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void)
{
    mT2ClearIntFlag();
    if(selectSnare == 1){
        if(NumISR == 0){
            i = index >> 1;
            if (~(index & 1))
                packed = Snare[i];
            if (index & 1)
                DAC_input = packed >> 4;   //set the upper 4 bits
            else 
                DAC_input = packed & 0x0f;//set the lower 4 bits
            index++;
            if(i > upperIndex || i > sizeof(Snare))
                selectSnare = 0;
        }
        //ISR Triggered every 780 timer count; But the output needs to be 16 KHz,
        //which is 2500 timer count. So, the sample needs te played for four cycles
        NumISR = NumISR + 1;
        if(NumISR == 4){
            NumISR = 0;
        }
        mPORTBClearBits(BIT_4);
        while (TxBufFullSPI2());
        WriteSPI2(DAC_config_chan_A | (DAC_input << 8));
        while (SPI2STATbits.SPIBUSY);
        mPORTBSetBits(BIT_4);
    }
     
    else if(selectFM){
        phase_accum_fm += phase_incr_fm; 
        // main phase
        phase_accum_main += phase_incr_main + multfix16(sine_table[phase_accum_fm>>24], env_fm);
        // init the exponential decays by adding energy to the exponential filters 
        if (playIndiv || playBack) {
            dk_state_fm = fm_depth; 
            dk_state_main = onefix16; 
            attack_state_fm = fm_depth; 
            attack_state_main = onefix16; 
            playIndiv = 0; 
            phase_accum_fm = 0;
            phase_accum_main = 0;
            dk_interval = 0;
            sustain_state = 0;
        }
        // envelope calculations are 256 times slower than sample rate
        // computes 4 exponential decays and builds the product envelopes
        if ((dk_interval++ & 0xff) == 0){
            dk_state_fm = multfix16(dk_state_fm, dk_fm) ;
            dk_state_main = multfix16(dk_state_main, dk_main) ;
            attack_state_fm = multfix16(attack_state_fm, attack_fm);
            attack_state_main = multfix16(attack_state_main, attack_main);
            env_fm = multfix16(fm_depth-attack_state_fm, dk_state_fm) ;
            env_main = multfix16(onefix16-attack_state_main, dk_state_main) ;
            if(sustain_state < sustain_interval) {
                dk_state_main = onefix16;
                sustain_state = sustain_state + sustain_constant ;
            }
        }
        // === DAC Output === //
        // Channel A
        // CS low to start transaction
        mPORTBClearBits(BIT_4); // start transaction
        WriteSPI2(DAC_data);     
        wave_main = multfix16(sine_table[phase_accum_main>>24], env_main);
        // truncate to 12 bits, read table, convert to int and add offset
        DAC_data = DAC_config_chan_A | (fix2int16(wave_main) + 2048) ; 
        // test for done
        while (SPI2STATbits.SPIBUSY); // wait for end of transaction
         // CS high
        mPORTBSetBits(BIT_4); // end transaction
        isr_time = max(ReadTimer2(), isr_time) ; // - isr_time;
    }
    else if (selectKS){
        previdx = idx;
        idx = idx + 1;
        randval = rand()%10;
        switch (key){
            case 1:
                sample = stringC[idx%C_length];
                stringC[idx%C_length] = (stringC[idx%C_length] + stringC[previdx%C_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < C_length; k++){
                        stringC[k] = random_table[k];
                    }
                }
                break;
            case 2:
                sample = stringD[idx%D_length];
                stringD[idx%D_length] = (stringD[idx%D_length] + stringD[previdx%D_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < D_length; k++){
                        stringD[k] = random_table[k];
                    }
                }
                break;
            case 3:
                sample = stringE[idx%E_length];
                stringE[idx%E_length] = (stringE[idx%E_length] + stringE[previdx%E_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < E_length; k++){
                        stringE[k] = random_table[k];
                    }
                }
                break;
            case 4:
                sample = stringF[idx%F_length];
                stringF[idx%F_length] = (stringF[idx%F_length] + stringF[previdx%F_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < F_length; k++){
                        stringF[k] = random_table[k];
                    }
                }
                break;
            case 5: 
                sample = stringG[idx%G_length];
                stringG[idx%G_length] = (stringG[idx%G_length] + stringG[previdx%G_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < G_length; k++){
                        stringG[k] = random_table[k];
                    }
                }
                break;
            case 6:
                sample = stringA[idx%A_length];
                stringA[idx%A_length] = (stringA[idx%A_length] + stringA[previdx%A_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < A_length; k++){
                        stringA[k] = random_table[k];
                    }
                }
                break;
            case 7:
                sample = stringB[idx%B_length];
                stringB[idx%B_length] = (stringB[idx%B_length] + stringB[previdx%B_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < B_length; k++){
                        stringB[k] = random_table[k];
                    }
                }
                break;
            case 8:
                sample = stringC5[idx%C5_length];
                stringC5[idx%C5_length] = (stringC5[idx%C5_length] + stringC5[previdx%C5_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < C5_length; k++){
                        stringC5[k] = random_table[k];
                    }
                }
                break;
            case 9:
                sample = stringD5[idx%D5_length];
                stringD5[idx%D5_length] = (stringD5[idx%D5_length] + stringD5[previdx%D5_length])/2;
                if(idx > 4000){
                    selectKS = 0;
                    idx = 0;
                    for(k = 0; k < D5_length; k++){
                        stringD5[k] = random_table[k];
                    }
                }
                break;
            default:
                if(idx > 4000){
                    selectKS = 0;
                }
                break;
        }
        mPORTBClearBits(BIT_4); // start transaction
        WriteSPI2(DAC_data);     
        // truncate to 12 bits, read table, convert to int and add offset
        DAC_data = DAC_config_chan_A | (sample + 2047) ; 
        // test for done
        while (SPI2STATbits.SPIBUSY); // wait for end of transaction
         // CS high
        mPORTBSetBits(BIT_4); // end transaction
    }
} // end ISR TIMER2

void home_page(){
    tft_fillScreen(ILI9340_BLACK);
    tft_setTextSize(3);
    tft_setCursor(70, 5);
    sprintf(display, "Music");
    tft_writeString(display);
    tft_setCursor(40, 30);   
    sprintf(display, "Sequencer");
    tft_writeString(display);           
    tft_setCursor(50, 150);
    sprintf(display, "* START");
    tft_writeString(display);
    tft_setCursor(50, 180);
    sprintf(display, "0 HELP");
    tft_writeString(display);
    tft_setCursor(50, 210);
    sprintf(display, "# RESET");
    tft_writeString(display);
}

void help_page(int page){
    tft_fillScreen(ILI9340_BLACK);
    if (page==0){
        tft_setTextSize(2);
        tft_setCursor(80, 3);
        sprintf(display, "Help");
        tft_writeString(display);
        tft_setTextSize(1);
        tft_setCursor(5, 20);
        sprintf(display, "Press keys 1-9 to listen to the sound:");
        tft_writeString(display);
        tft_setCursor(5, 30);
        sprintf(display, "1-Snare Drum 2-           3-           ");
        tft_writeString(display);
        tft_setCursor(5, 40);
        sprintf(display, "4-           5-           6-           ");
        tft_writeString(display);
        tft_setCursor(5, 50);
        sprintf(display, "7-           8-           9-           ");
        tft_writeString(display);    
        tft_setCursor(5, 65);
        sprintf(display, "Press keys A-D to select a mode. Press");
        tft_writeString(display); 
        tft_setCursor(5, 75);
        sprintf(display, "the same button to exit current mode.");
        tft_writeString(display); 
        tft_setCursor(5, 85);
        sprintf(display, "A-Record Mode B-Playback Mode");
        tft_writeString(display); 
        tft_setCursor(5, 95);
        sprintf(display, "C-Edit Mode   D-Delete Mode");
        tft_writeString(display); 
        tft_setCursor(5, 110);
        sprintf(display, "After pressing A for record mode, press");
        tft_writeString(display); 
        tft_setCursor(5, 120);
        sprintf(display, "any combinations of buttons 1-9 (once");
        tft_writeString(display);
        tft_setCursor(5, 130);   
        sprintf(display, "per button). Based on your selections,");
        tft_writeString(display);    
        tft_setCursor(5, 140);   
        sprintf(display, "a set of music patterns will be");
        tft_writeString(display);   
        tft_setCursor(5, 150);   
        sprintf(display, "generated and stored (Up to 14");
        tft_writeString(display);   
        tft_setCursor(5, 160);   
        sprintf(display, " different sets may be stored).");
        tft_writeString(display);   
        tft_setCursor(5, 175);   
        sprintf(display, "After pressing B for playback mode, a");
        tft_writeString(display);   
        tft_setCursor(5, 185);   
        sprintf(display, "list of stored pattern sets will be");
        tft_writeString(display);   
        tft_setCursor(5, 195);   
        sprintf(display, "shown. Use the corresponding keys to");
        tft_writeString(display);       
        tft_setCursor(5, 205);   
        sprintf(display, "select the set you want to listen to,");
        tft_writeString(display);
        tft_setCursor(5, 215);   
        sprintf(display, "the default pattern will be played.");
        tft_writeString(display);   
        tft_setCursor(5, 225);   
        sprintf(display, "To change the default pattern, use the");
        tft_writeString(display);   
        tft_setCursor(5, 235);   
        sprintf(display, "edit mode instead.");
        tft_writeString(display);
        tft_setCursor(170, 300);   
        sprintf(display, "# Next page");
        tft_writeString(display);
    }
    else {
        tft_setCursor(5, 5);   
        sprintf(display, "After pressing C for edit mode, a");
        tft_writeString(display);   
        tft_setCursor(5, 15);   
        sprintf(display, "list of stored pattern sets will be");
        tft_writeString(display);   
        tft_setCursor(5, 25);   
        sprintf(display, "shown. Use the corresponding keys to");
        tft_writeString(display);       
        tft_setCursor(5, 35);   
        sprintf(display, "select the set you want to edit. Press");
        tft_writeString(display);
        tft_setCursor(5, 45);   
        sprintf(display, "the corresponding keys to iterate");
        tft_writeString(display);   
        tft_setCursor(5, 55);   
        sprintf(display, "through the set of possible patterns.");
        tft_writeString(display);
        tft_setCursor(5, 70);   
        sprintf(display, "After pressing D for delete mode,");
        tft_writeString(display);       
        tft_setCursor(5, 80);   
        sprintf(display, "a list of stored pattern sets will");
        tft_writeString(display);
        tft_setCursor(5, 90);   
        sprintf(display, "be shown. Use the corresponding keys");
        tft_writeString(display);
        tft_setCursor(5, 100);   
        sprintf(display, "to select the set you want to delete.");
        tft_writeString(display);
        tft_setCursor(5, 120);   
        sprintf(display, "Press keys *,0,# to navigate the home");
        tft_writeString(display);       
        tft_setCursor(5, 130);   
        sprintf(display, "page:");
        tft_writeString(display);
        tft_setCursor(5, 140);   
        sprintf(display, "*-Start using the music sequencer");
        tft_writeString(display);
        tft_setCursor(5, 150);   
        sprintf(display, "0-Help on using the music sequencer");
        tft_writeString(display);
        tft_setCursor(5,160);
        sprintf(display, "#-Reset and clear stored pattern sets");
        tft_writeString(display);
        tft_setCursor(5, 300);   
        sprintf(display, "* Previous page");
        tft_writeString(display);        
    }
    tft_setCursor(50, 280);   
    sprintf(display, "0 Return to home page");
    tft_writeString(display);
}

void reset_page() {
    tft_fillScreen(ILI9340_BLACK);
    tft_setTextSize(2);
    tft_setCursor(30, 150);
    sprintf(display, "Memory cleared!");
    tft_writeString(display);
    tft_setTextSize(1);
    tft_setCursor(50, 280);   
    sprintf(display, "# Return to home page");
    tft_writeString(display);
}

void fill_playString(int* keystring, int size){
    int k, randnum;
    for(k = 0; k < 50; k++){
        randnum = rand()%size;
        playString[k] = keystring[randnum];
    }
}

//Protothread for maintaining ISR-driven millisecond-timing
static PT_THREAD (protothread_keypad(struct pt *pt)){
    PT_BEGIN(pt);
    static int keyPad, press, recordMode, instrMode, i, keyInput, playIndex = 0, j;
    static int pushState = release, keyIndex = 0, portA = 0, started = 0, help = 0, reset=0, home=0;
    //Port A ports as digital output
    mPORTASetPinsDigitalOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);
    //Port B ports as digital input
    mPORTBSetPinsDigitalIn(BIT_5 | BIT_7 | BIT_8 | BIT_9);
    tft_setTextColor(ILI9340_WHITE);
    tft_setTextSize(1);
    tft_setCursor(5, 5);
    while(1){
        //On start up, no keys pressed
        if (started == 0 && help == 0 && reset == 0 && home==0) {
            home=1;
            home_page();
        }
        mPORTAClearBits(BIT_0 | BIT_1 | BIT_2 | BIT_3);
        portA = 1;
        mPORTASetBits(portA);
        PT_YIELD_TIME_msec(30);
        //get valid key input
        for(i = 0; i < 4; i++){
            //Read the port B value
            keyInput = mPORTBReadBits(BIT_5 | BIT_7 | BIT_8 | BIT_9);
            //If there is a keypress in the search row
            if(keyInput != 0){
                keyInput |= portA;
                break;
            }
            mPORTAClearBits(portA);
            portA <<= 1;
            mPORTASetBits(portA);
        }
        //check if the generated keycode is valid or not
        if(keyInput > 0){
            for(i = 0; i < 16; i++){
                if(keyInput == keytable[i]){
                    break;
                }
            }
        }
        else 
            i = -1; 
        keyPad = i;
        //compute the key-press state
        if (keyPad != -1)
            press = 1;
        else
            press = 0;       
       //Setup the finite state machine
        switch(pushState){
            case release:
                if(press){
                    pushState = mayPressed;
                }
                else{
                    pushState = release;
                }
                break;
            case mayPressed:
                if(press){
                    pushState = pressed;
                    if (started) { 
                        if (keyPad==10 && recordMode == 0) { //'A' pressed -- Record mode
                            recordMode = 1;
                            tft_setTextSize(1);
                            tft_setCursor(100, 100);
                            sprintf(display, "In Record Mode");
                            tft_writeString(display);
                        }
                        else if (keyPad==13 && instrMode == 0) { //'A' pressed -- Record mode
                            instrMode = 1;
                            tft_setTextSize(1);
                            tft_setCursor(100, 100);
                            sprintf(display, "In instrument Mode");
                            tft_writeString(display);
                        }
                        else if (instrMode == 1){                        
                            instrString[playIndex] = keyPad;
                            playIndex ++;
                            switch (keyPad) {
                                case 1:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 2:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 3:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 4:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 5:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 6:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 7:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 8:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 9:
                                    selectKS = 1;
                                    key = keyPad;
                                    break;
                                case 13:
                                   tft_setTextSize(1);
                                   tft_setCursor(100, 200);
                                   sprintf(display, "ended instrument mode");
                                   tft_writeString(display);
                                   instrMode = 0;
                                   selectKS = 0;
                                   break;
                                default:
                                   break;
                            }
                        }
                        else if (recordMode == 1) {
                            keyString[keyIndex] = keyPad;
                            keyIndex ++;
                            switch (keyPad) {
                                case 1: //snare drum -- wave table
                                    selectKS = 0;
                                    selectFM = 0;
                                    selectSnare = 1;
                                    index = 0;
                                    upperIndex = 1300;
                                    break;
                                case 2: //lazer gun -- FM
                                    selectFM = 1;
                                    playIndiv=1;
                                    phase_incr_main = 100.0*(float)two32/Fs;
                                    phase_incr_fm = 100.0*(float)two32/Fs;
                                    attack_main = float2fix16(0);
                                    attack_fm = float2fix16(0);
                                    dk_main = float2fix16(0.87);
                                    dk_fm = float2fix16(0.88);
                                    fm_depth = float2fix16(4);
                                    break;
                                case 3: //bass drum -- FM
                                    playIndiv=1;
                                    selectFM = 1;
                                    phase_incr_main = 50.0*(float)two32/Fs;
                                    phase_incr_fm = 90.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.97);
                                    dk_fm = float2fix16(0.98);
                                    fm_depth = float2fix16(1.0);
                                    break;                             
                                case 4: //triangle -- FM
                                    playIndiv = 1;
                                    selectFM = 1;
                                    phase_incr_main = 2000.0*(float)two32/Fs;
                                    phase_incr_fm = 3000.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.97);
                                    dk_fm = float2fix16(0.98);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 5: //cowbell -- FM
                                    playIndiv = 1;
                                    selectFM = 1;
                                    phase_incr_main = 800.0*(float)two32/Fs;
                                    phase_incr_fm = 1400.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.97);
                                    dk_fm = float2fix16(0.98);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 6: //hit-hat --FM 
                                    playIndiv=1;
                                    selectFM = 1;
                                    phase_incr_main = 213.0*(float)two32/Fs;
                                    phase_incr_fm = 713.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.5);
                                    dk_fm = float2fix16(0.88);
                                    fm_depth = float2fix16(5);
                                    break;
                                case 7: //bounce ball --FM
                                    selectFM = 1;
                                    playIndiv=1;
                                    phase_incr_main = 130.0*(float)two32/Fs;
                                    phase_incr_fm = 234.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.99);
                                    dk_fm = float2fix16(0.97);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 8: //wood block --FM
                                    selectFM = 1;
                                    playIndiv=1;
                                    phase_incr_main = 123.0*(float)two32/Fs;
                                    phase_incr_fm = 234.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.9);
                                    dk_fm = float2fix16(0.89);
                                    fm_depth = float2fix16(1.0);
                                    break;
                                case 9: //temple bell -- FM
                                    selectFM = 1;
                                    playIndiv=1;
                                    phase_incr_main = 231.0*(float)two32/Fs;
                                    phase_incr_fm = 734.0*(float)two32/Fs;
                                    attack_main = float2fix16(0);
                                    attack_fm = float2fix16(0);
                                    dk_main = float2fix16(0.98);
                                    dk_fm = float2fix16(0.99);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 10://end record mode
                                    tft_setTextSize(1);
                                    tft_setCursor(100, 200);
                                    sprintf(display, "ended record mode");
                                    tft_writeString(display);
                                    recordMode = 0;
                                    break;
                                default: //do nothing since cannot be non-instrument keys in record mode
                                    break;
                            }
                        }
                        else if (recordMode == 0) {
                            switch (keyPad) {
                                case 1: //snare drum -- WAVE TABLE
                                    selectKS = 0;
                                    selectFM = 0;
                                    selectSnare = 1;
                                    index = 0;
                                    upperIndex = 1300;
                                    break;
                                case 2:
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    playIndiv=1;
                                    phase_incr_main = 100.0*(float)two32/Fs;
                                    phase_incr_fm = 100.0*(float)two32/Fs;
                                    attack_main = float2fix16(0);
                                    attack_fm = float2fix16(0);
                                    dk_main = float2fix16(0.87);
                                    dk_fm = float2fix16(0.88);
                                    fm_depth = float2fix16(4);
                                    break;
                                case 3: //bass drum -- FM
                                    playIndiv=1;
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    phase_incr_main = 50.0*(float)two32/Fs;
                                    phase_incr_fm = 90.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.97);
                                    dk_fm = float2fix16(0.98);
                                    fm_depth = float2fix16(0.5);
                                    break;                             
                                case 4: //triangle -- FM
                                    playIndiv = 1;
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    phase_incr_main = 2000.0*(float)two32/Fs;
                                    phase_incr_fm = 3000.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.97);
                                    dk_fm = float2fix16(0.98);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 5: //cowbell -- FM
                                    playIndiv = 1;
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    phase_incr_main = 800.0*(float)two32/Fs;
                                    phase_incr_fm = 1400.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.97);
                                    dk_fm = float2fix16(0.98);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 6: //hit-hat
                                    playIndiv=1;
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    phase_incr_main = 213.0*(float)two32/Fs;
                                    phase_incr_fm = 713.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.5);
                                    dk_fm = float2fix16(0.88);
                                    fm_depth = float2fix16(5);
                                    break;
                                case 7:
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    playIndiv=1;
                                    phase_incr_main = 130.0*(float)two32/Fs;
                                    phase_incr_fm = 234.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.99);
                                    dk_fm = float2fix16(0.97);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 8:
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    playIndiv=1;
                                    phase_incr_main = 123.0*(float)two32/Fs;
                                    phase_incr_fm = 234.0*(float)two32/Fs;
                                    attack_main = float2fix16(0.0);
                                    attack_fm = float2fix16(0.0);
                                    dk_main = float2fix16(0.9);
                                    dk_fm = float2fix16(0.89);
                                    fm_depth = float2fix16(1.0);
                                    break;
                                case 9:
                                    selectKS = 0;
                                    selectFM = 1;
                                    selectSnare = 0;
                                    playIndiv=1;
                                    phase_incr_main = 231.0*(float)two32/Fs;
                                    phase_incr_fm = 734.0*(float)two32/Fs;
                                    attack_main = float2fix16(0);
                                    attack_fm = float2fix16(0);
                                    dk_main = float2fix16(0.98);
                                    dk_fm = float2fix16(0.99);
                                    fm_depth = float2fix16(2.5);
                                    break;
                                case 10:
                                    break;
                                case 11:
                                    if(keyIndex != 0){
                                        fill_playString(keyString, keyIndex);
                                        for(j = 0; j < 50; j ++){
                                            if(playString[j] == 1){
                                                selectKS = 0;
                                                selectFM = 0;
                                                selectSnare = 1;
                                                index = 0;
                                                upperIndex = 1300;
                                            }
                                            else if(playString[j] == 2){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                playIndiv=1;
                                                phase_incr_main = 100.0*(float)two32/Fs;
                                                phase_incr_fm = 100.0*(float)two32/Fs;
                                                attack_main = float2fix16(0);
                                                attack_fm = float2fix16(0);
                                                dk_main = float2fix16(0.87);
                                                dk_fm = float2fix16(0.88);
                                                fm_depth = float2fix16(4);
                                            }
                                            else if(playString[j] == 3){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                playIndiv=1;
                                                phase_incr_main = 50.0*(float)two32/Fs;
                                                phase_incr_fm = 90.0*(float)two32/Fs;
                                                attack_main = float2fix16(0.0);
                                                attack_fm = float2fix16(0.0);
                                                dk_main = float2fix16(0.97);
                                                dk_fm = float2fix16(0.98);
                                                fm_depth = float2fix16(0.5);
                                            }
                                            else if(playString[j] == 4){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                playIndiv=1;
                                                phase_incr_main = 2000.0*(float)two32/Fs;
                                                phase_incr_fm = 3000.0*(float)two32/Fs;
                                                attack_main = float2fix16(0.0);
                                                attack_fm = float2fix16(0.0);
                                                dk_main = float2fix16(0.97);
                                                dk_fm = float2fix16(0.98);
                                                fm_depth = float2fix16(2.5);
                                            }
                                            else if(playString[j] == 5){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                playIndiv=1;
                                                phase_incr_main = 800.0*(float)two32/Fs;
                                                phase_incr_fm = 1400.0*(float)two32/Fs;
                                                attack_main = float2fix16(0.0);
                                                attack_fm = float2fix16(0.0);
                                                dk_main = float2fix16(0.97);
                                                dk_fm = float2fix16(0.98);
                                                fm_depth = float2fix16(2.5);
                                            }
                                            else if(playString[j] == 6){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                phase_incr_main = 213.0*(float)two32/Fs;
                                                phase_incr_fm = 713.0*(float)two32/Fs;
                                                attack_main = float2fix16(0.0);
                                                attack_fm = float2fix16(0.0);
                                                dk_main = float2fix16(0.5);
                                                dk_fm = float2fix16(0.88);
                                                fm_depth = float2fix16(5);
                                            }
                                            else if(playString[j] == 7){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                playIndiv=1;
                                                phase_incr_main = 130.0*(float)two32/Fs;
                                                phase_incr_fm = 234.0*(float)two32/Fs;
                                                attack_main = float2fix16(0.0);
                                                attack_fm = float2fix16(0.0);
                                                dk_main = float2fix16(0.99);
                                                dk_fm = float2fix16(0.97);
                                                fm_depth = float2fix16(2.5);
                                            }
                                            else if(playString[j] == 8){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                playIndiv=1;
                                                phase_incr_main = 123.0*(float)two32/Fs;
                                                phase_incr_fm = 234.0*(float)two32/Fs;
                                                attack_main = float2fix16(0.0);
                                                attack_fm = float2fix16(0.0);
                                                dk_main = float2fix16(0.9);
                                                dk_fm = float2fix16(0.89);
                                                fm_depth = float2fix16(0);
                                            }
                                            else if(playString[j] == 9){
                                                selectKS = 0;
                                                selectFM = 1;
                                                selectSnare = 0;
                                                playIndiv=1;
                                                phase_incr_main = 231.0*(float)two32/Fs;
                                                phase_incr_fm = 734.0*(float)two32/Fs;
                                                attack_main = float2fix16(0);
                                                attack_fm = float2fix16(0);
                                                dk_main = float2fix16(0.98);
                                                dk_fm = float2fix16(0.99);
                                                fm_depth = float2fix16(2.5);
                                            }
                                            else{
                                                playBack=0;
                                            }
                                            playBack=1;
                                            PT_YIELD_TIME_msec(300);
                                            playBack = 0;
                                            PT_YIELD_TIME_msec(80);
                                        }
                                    }
                                    break;
                                case 12:
                                    if(playIndex != 0){
                                        fill_playString(instrString, playIndex);
                                        for(j = 0; j < 50; j ++){
                                            if(playString[j] == 1){
                                                selectKS = 1;
                                                key = 1;
                                            }
                                            else if(playString[j] == 2){
                                                selectKS = 1;
                                                key = 2;
                                            }
                                            else if(playString[j] == 3){
                                                selectKS = 1;
                                                key = 3;
                                            }
                                            else if(playString[j] == 4){
                                                selectKS = 1;
                                                key = 4;
                                            }
                                            else if(playString[j] == 5){
                                                selectKS = 1;
                                                key = 5;
                                            }
                                            else if(playString[j] == 6){
                                                selectKS = 1;
                                                key = 6;
                                            }
                                            else if(playString[j] == 7){
                                                selectKS = 1;
                                                key = 7;
                                            }
                                            else if(playString[j] == 8){
                                                selectKS = 1;
                                                key = 8;
                                            }
                                            else if(playString[j] == 9){
                                                selectKS = 1;
                                                key = 9;
                                            }
                                            else{
                                                playBack=0;
                                            }
                                            playBack=1;
                                            PT_YIELD_TIME_msec(200);
                                            playBack = 0;
                                            PT_YIELD_TIME_msec(50);
                                        }
                                    }
                                    break;
                                default:
                                    break;
                            }
                        }
                        else if (keyPad==14) {
                            started=0;
                            home_page();
                        }

                    }
                    else if (keyPad==14 && !reset) { //* pressed
                        if (help) { //* pressed at help page to go to previous page
                            help_page(0);
                        }
                        else { //*pressed at home page to start
                            started = 1;
                            tft_setTextSize(1);
                            tft_setCursor(5,5);
                            tft_fillScreen(ILI9340_BLACK);
                            sprintf(display, "Music Sequence Started");
                            tft_writeString(display);
                        }
                    }
                    else if (keyPad==0 && !reset) { //0 pressed -- Help
                        if (!help) { //0 pressed at home page to enter help page
                            help = 1;
                            help_page(0);
                        }
                        else { //0 pressed at help page to exit page
                            help = 0;
                            home_page();
                        }
                    }
                    else if (keyPad==15) { //# pressed
                        if (help) { //# pressed at help page to go to next page
                            help_page(1);
                        }
                        else if (reset){ //# pressed at reset page to exit page
                            reset = 0;
                            home_page();
                        }
                        else { //# pressed at home page
                            reset = 1;
                            reset_page();
                        }
                    }
                }
                else
                    pushState = release;
                break;
            case pressed:
                if(press){
                    pushState = pressed;
                }
                else pushState = mayRelease;
                break;
            case mayRelease:
                if(press){
                    pushState = pressed;
                }
                else pushState = release;
                break;
        }
    }
    PT_END(pt);
}

static PT_THREAD (protothread_tick(struct pt *pt)){
    PT_BEGIN(pt);
    static int sys_time_seconds = 0;
    while(1){
          // yield time 1 second
          PT_YIELD_TIME_msec(1000);
          sys_time_seconds = sys_time_seconds + 1;
          // NEVER exit while
    } // END WHILE(1)
    PT_END(pt);
}
    
int main(void){
    // === config the uart, DMA, vref, timer5
    SYSTEMConfigPerformance(PBCLK);
    ANSELA = 0; ANSELB = 0; CM1CON = 0; CM2CON = 0;
    //for DAC transaction 
    PT_setup();    
    //set up the timer
    //sample rate = 20k
    OpenTimer2(T2_ON| T2_SOURCE_INT | T2_PS_1_1, 2000);
    ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    mT2ClearIntFlag(); //clear all pre-existing flags 
    
    //set up the SPI and DAC   
    PPSOutput(3, RPA4, SDO2); //MOSI output at pin 12
    
    // control CS for DAC
    mPORTBSetPinsDigitalOut(BIT_4);
    mPORTBSetBits(BIT_4);
    SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV, spiClkDiv);    
    
    //Init tft display
    tft_init_hw();
    tft_begin();
    tft_fillScreen(ILI9340_BLACK);
    tft_setRotation(0);
    
    int i;
    //sine table for FM synthesis
    for(i = 0; i < sine_table_size; i++){
        sine_table[i] = float2fix16(2047*sin((float)i*6.283/(float)sine_table_size));
        random_table[i] = rand()%2047;
        stringC[i%C_length] = random_table[i];
        stringD[i%D_length] = random_table[i];
        stringE[i%E_length] = random_table[i];
        stringF[i%F_length] = random_table[i];
        stringG[i%G_length] = random_table[i];
        stringA[i%A_length] = random_table[i];
        stringB[i%B_length] = random_table[i];
    }
    
    // === setup global interrupt
    INTEnableSystemMultiVectoredInt();
    
    //init the threads
    PT_INIT(&pt_keypad);
    PT_INIT(&pt_tick);
    
    while(1){
        PT_SCHEDULE(protothread_keypad(&pt_keypad));
        PT_SCHEDULE(protothread_tick(&pt_tick));
    }
}